import { ArrayList, HashMap } from '@kit.ArkTS';
import http from '@ohos.net.http';
import { SZYXLocalStorageHelper } from '../utils/SZYXLocalStorageHelper';
import { SZYXLocalStorageKeys } from '../utils/SZYXLocalStorageKeys';
import { ToolsHelperForTS } from '../utils/ToolsHelperForTS';


type HttpParamsGet = {
  url: string
  query?: Record<string, string> | Object
  headers?: Record<string, string>
}
type HttpParamsPost = {
  url: string
  data?: string | Object | ArrayBuffer
  query?: Record<string, string> | Object
  headers?: Record<string, string>
}
type HttpParamsForm = {
  url: string
  data?: Record<string, string> | Object|undefined
  query?: Record<string, string> | Object
  headers?: Record<string, string>
}

export class HttpHelper {
  private static instance: HttpHelper | null = null

  // 单例模式
  static get() {
    // 判断系统是否已经有单例了
    if (HttpHelper.instance === null) {
      HttpHelper.instance = new HttpHelper()
    }
    return HttpHelper.instance
  }

  //请求中队列
  private httpHandlerList = new HashMap<string, http.HttpRequest>();
  // 并发白名单，这个名单里面的api，重复请求不会取消
  private concurrentList = new ArrayList<string>();

  constructor() {
    this.httpHandlerList = new HashMap<string, http.HttpRequest>();
    SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerList, this.httpHandlerList)
    SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerListLength, this.httpHandlerList.length)
    this.concurrentList.clear()
  }

  /**
   * 添加并发白名单
   * @param apiNo
   */
  public addConcurrent(apiNo: string) {
    if (this.concurrentList.getIndexOf(apiNo) === -1) {
      this.concurrentList.add(apiNo)
    }
  }

  public removeConcurrent(apiNo: string) {
    if (this.concurrentList.getIndexOf(apiNo) !== -1) {
      this.concurrentList.remove(apiNo)
    }
  }

  /**
   * postJson请求
   * @param url url地址
   * @param headers
   * @param apiNo 请求标识，取消请求或者去重使用|考虑做自动重试使用
   * @returns
   */
  public postJson<T>(params: HttpParamsPost, apiNo?: string): Promise<T> {

    return new Promise<T>((resolve, reject) => {

      if (this.concurrentList.getIndexOf(apiNo ?? params.url) === -1 &&
      this.httpHandlerList.hasKey(apiNo ?? params.url)) {
        this.httpHandlerList.get(apiNo ?? params.url).destroy()
        this.httpHandlerList.remove(apiNo ?? params.url)
        SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerList, this.httpHandlerList)
        SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerListLength, this.httpHandlerList.length)
      }
      let httpRequest = http.createHttp();

      if (this.concurrentList.getIndexOf(apiNo ?? params.url) === -1) {
        this.httpHandlerList.set(apiNo ?? params.url, httpRequest)
        SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerList, this.httpHandlerList)
        SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerListLength, this.httpHandlerList.length)
      }

      const header = {
        "Content-Type": "application/json;charset=UTF-8",
        // "Accept": "application/json",
        ...params.headers
      }
      console.log('=====>', 'POST:', JSON.stringify(params))
      // console.log('=====>', '接口请求', JSON.stringify(header))
      // console.log('=====>', '接口请求', data)

      httpRequest.request(this.getUrl(params.url, params.query), {
        method: http.RequestMethod.POST,
        connectTimeout: 20000,
        readTimeout: 20000,
        header: header,
        extraData: params.data
      })
        .then((data: http.HttpResponse) => {
          console.info(`=====>Result:${data.result as string}(${apiNo})`);
          // console.info('=====>' + 'code:' + data.responseCode);
          // console.info('=====>' + 'type:' + JSON.stringify(data.resultType));
          // console.info('=====>' + 'header:' + JSON.stringify(data.header));
          // console.info('=====>' + 'cookies:' + data.cookies); // 自API version 8开始支持cookie
          // console.info('=====>' + 'header.content-Type:' + JSON.stringify(data.header));
          // console.info('=====>' + 'header.Status-Line:' + JSON.stringify(data.header));

          if (this.httpHandlerList.hasKey(apiNo ?? params.url)) {
            this.httpHandlerList.remove(apiNo ?? params.url)
            SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerList, this.httpHandlerList)
            SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerListLength, this.httpHandlerList.length)
          }
          if (data.responseCode === 200) {
            resolve((typeof data.result === 'string' ? JSON.parse(data.result) : data.result) as T)
          } else {
            const err: Error = new Error()
            err.name = data.responseCode.toString()
            err.message = '服务异常'
            reject(err)
          }
        }).catch((err: Error) => {
        console.info('=====>' + 'Error:' + JSON.stringify({ err: err, url: params.url, }));
        if (this.httpHandlerList.hasKey(apiNo ?? params.url)) {
          this.httpHandlerList.remove(apiNo ?? params.url)
          SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerList, this.httpHandlerList)
          SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerListLength, this.httpHandlerList.length)
        }
        if (err.message === 'Failed writing received data to disk/application') {
          reject('cancel')
        } else {
          reject(err)
        }
      });
    });

  }

  /**
   * postForm请求
   * @param url url地址
   * @param headers
   * @param apiNo 请求标识，取消请求或者去重使用|考虑做自动重试使用
   * @returns
   */
  public postForm<T>(params: HttpParamsForm, apiNo?: string): Promise<T> {

    return new Promise<T>((resolve, reject) => {

      if (this.concurrentList.getIndexOf(apiNo ?? params.url) === -1 &&
      this.httpHandlerList.hasKey(apiNo ?? params.url)) {
        this.httpHandlerList.get(apiNo ?? params.url).destroy()
        this.httpHandlerList.remove(apiNo ?? params.url)
        SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerList, this.httpHandlerList)
        SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerListLength, this.httpHandlerList.length)
      }
      let httpRequest = http.createHttp();

      if (this.concurrentList.getIndexOf(apiNo ?? params.url) === -1) {
        this.httpHandlerList.set(apiNo ?? params.url, httpRequest)
        SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerList, this.httpHandlerList)
        SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerListLength, this.httpHandlerList.length)
      }

      const header = {
        "Content-Type": "application/x-www-form-urlencoded;charset=UTF-8",
        // "Accept": "application/json",
        ...params.headers
      }
      let data = this.getContent(params.data)

      console.log('=====>', 'POSTForm:', params.url)
      console.log('=====>', 'POSTForm:', JSON.stringify(header))
      console.log('=====>', 'POSTForm:', data)

      httpRequest.request(this.getUrl(params.url, params.query), {
        method: http.RequestMethod.POST,
        connectTimeout: 20000,
        readTimeout: 20000,
        header: header,
        extraData: data?encodeURI(data):undefined
      })
        .then((data: http.HttpResponse) => {
          console.info(`=====>Result:${data.result as string}(${apiNo})`);
          // console.info('=====>' + 'code:' + data.responseCode);
          // console.info('=====>' + 'type:' + JSON.stringify(data.resultType));
          // console.info('=====>' + 'header:' + JSON.stringify(data.header));
          // console.info('=====>' + 'cookies:' + data.cookies); // 自API version 8开始支持cookie
          // console.info('=====>' + 'header.content-Type:' + JSON.stringify(data.header));
          // console.info('=====>' + 'header.Status-Line:' + JSON.stringify(data.header));

          if (this.httpHandlerList.hasKey(apiNo ?? params.url)) {
            this.httpHandlerList.remove(apiNo ?? params.url)
            SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerList, this.httpHandlerList)
            SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerListLength, this.httpHandlerList.length)
          }
          if (data.responseCode === 200) {
            resolve((typeof data.result === 'string' ? JSON.parse(data.result) : data.result) as T)
          } else {
            const err: Error = new Error()
            err.name = data.responseCode.toString()
            err.message = '服务异常'
            reject(err)
          }
        }).catch((err: Error) => {
        console.info('=====>' + 'Error:' + JSON.stringify({ err: err, url: params.url, }));
        if (this.httpHandlerList.hasKey(apiNo ?? params.url)) {
          this.httpHandlerList.remove(apiNo ?? params.url)
          SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerList, this.httpHandlerList)
          SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerListLength, this.httpHandlerList.length)
        }
        if (err.message === 'Failed writing received data to disk/application') {
          reject('cancel')
        } else {
          reject(err)
        }
      });
    });

  }


  /**
   * get请求
   * @param url url地址
   * @param data 请求参数
   * @param headers
   * @param apiNo 请求标识，取消请求或者去重使用|考虑做自动重试使用
   * @returns
   */
  public get<T>(params: HttpParamsGet, apiNo?: string): Promise<T> {

    return new Promise<T>((resolve, reject) => {

      if (this.concurrentList.getIndexOf(apiNo ?? params.url) === -1 &&
      this.httpHandlerList.hasKey(apiNo ?? params.url)) {
        this.httpHandlerList.get(apiNo ?? params.url).destroy()
        this.httpHandlerList.remove(apiNo ?? params.url)
        SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerList, this.httpHandlerList)
        SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerListLength, this.httpHandlerList.length)
      }
      let httpRequest = http.createHttp();

      if (this.concurrentList.getIndexOf(apiNo ?? params.url) === -1) {
        this.httpHandlerList.set(apiNo ?? params.url, httpRequest)
        SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerList, this.httpHandlerList)
        SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerListLength, this.httpHandlerList.length)
      }

      const header = {
        ...params.headers
      }
      console.log('=====>', 'GET:', this.getUrl(params.url, params.query))
      console.log('=====>', 'header:', JSON.stringify(header))

      httpRequest.request(this.getUrl(params.url, params.query), {
        method: http.RequestMethod.GET,
        connectTimeout: 20000,
        readTimeout: 20000,
        header: header,
        // extraData: params.data
      })
        .then((data: http.HttpResponse) => {
          console.info(`=====>Result:${data.result as string}(${apiNo})`);
          // console.info('=====>' + 'code:' + data.responseCode);
          // console.info('=====>' + 'type:' + JSON.stringify(data.resultType));
          // console.info('=====>' + 'header:' + JSON.stringify(data.header));
          // console.info('=====>' + 'cookies:' + data.cookies); // 自API version 8开始支持cookie
          // console.info('=====>' + 'header.content-Type:' + JSON.stringify(data.header));
          // console.info('=====>' + 'header.Status-Line:' + JSON.stringify(data.header));

          if (this.httpHandlerList.hasKey(apiNo ?? params.url)) {
            this.httpHandlerList.remove(apiNo ?? params.url)
            SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerList, this.httpHandlerList)
            SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerListLength, this.httpHandlerList.length)
          }
          if (data.responseCode === 200) {
            resolve((typeof data.result === 'string' ? JSON.parse(data.result) : data.result) as T)
          } else {
            const err: Error = new Error()
            err.name = data.responseCode.toString()
            err.message = '服务异常'
            reject(err)
          }
        }).catch((err: Error) => {
        console.info('=====>' + 'Error:' + JSON.stringify({ err: err, url: params.url, }));
        if (this.httpHandlerList.hasKey(apiNo ?? params.url)) {
          this.httpHandlerList.remove(apiNo ?? params.url)
          SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerList, this.httpHandlerList)
          SZYXLocalStorageHelper.storage.setOrCreate(SZYXLocalStorageKeys.HttpHandlerListLength, this.httpHandlerList.length)
        }
        if (err.message === 'Failed writing received data to disk/application') {
          reject('cancel')
        } else {
          reject(err)
        }
      });
    });

  }

  private getUrl(url: string, query?: Record<string, string> | Object) {
    let u = url
    if (query) {
      let q = query
      if (typeof query === 'object') {
        q = ToolsHelperForTS.classToRecord(query)
      }
      u = `${u}${u.indexOf('?') < 0 ? '?' : u.endsWith('$') ? '' : '&'}`
      Object.entries(q).forEach((row) => {
        if (row[1]) {
          u = `${u}${row[0]}=${row[1] as string}&`
        }
      });
      u = u.slice(0, u.length - 1)
    }
    return u
  }

  private getContent(data?: Record<string, string> | Object) {
    if (!data) {
      return undefined
    }
    let u = ''
    let q = data
    if (typeof data === 'object') {
      q = ToolsHelperForTS.classToRecord(data)
    }
    Object.entries(q).forEach((row) => {
      if (row[1]) {
        u = `${u}${row[0]}=${row[1] as string}&`
      }
    });
    u = u.slice(0, u.length - 1)
    return u
  }


}